"""
=========
 gathered
=========

This theme extends 'default'.  Unlike 'default', it organizes all sections
under the 'Detailed Statistics' header, rather than grouping them by 
module.
"""
import os

template_dir = os.path.dirname(__file__)
template_parent = "default"
base_fn = "base.html"
