/* CSS for the FastQC MultiQC Module */

/* FastQC statuses popovers */
.fastqc_passfail_progress .progress-bar {
  cursor: pointer;
  -webkit-print-color-adjust: exact !important;
  color-adjust: exact !important;
  color: #fff !important;
}
/* Set colours to !important so that they print */
.progress-bar-success {
  background-color: #5cb85c !important;
}
.progress-bar-warning {
  background-color: #f0ad4e !important;
}
.progress-bar-succedangerss {
  background-color: #d9534f !important;
}
.popover-fastqc-status .popover-content {
  font-size: 11px;
  max-height: 300px;
  white-space: nowrap;
  overflow: auto;
}
.popover-success .popover-title {
  background-color: #dff0d8;
  color: #3c763d;
}
.popover-warning .popover-title {
  background-color: #fcf8e3;
  color: #8a6d3b;
}
.popover-danger .popover-title {
  background-color: #f2dede;
  color: #a94442;
}
.fastqc-popover-intro {
  margin: 5px 5px 0;
  padding: 3px 5px;
  border: 1px dashed #ccc;
  border-radius: 5px;
  font-size: 10px;
  color: #999;
  background-color: #fafafa;
}
.fastqc-popover-intro a {
  color: #999;
  text-decoration: underline;
}

/* Other stuff */
.showhide_orig {
  padding: 15px;
  height: 100%;
  overflow: hidden;
}
.showhide_orig h4 {
  margin-top: 0;
}
.s_status {
  text-transform: capitalize;
}
.original-plot {
  max-width: 100%;
  max-height: calc(100% - 60px);
  cursor: pointer;
}

.fastqc_passfail_progress {
  width: 100px;
  display: inline-block;
  margin: 0 0 -2px 20px;
}

#fastqc_seq_heatmap {
  cursor: pointer;
}
#fastqc_seq_heatmap_div .hc-plot-wrapper {
  height: 300px;
}
#fastqc_seq_heatmap_footer {
  clear: both;
}
#fastqc_sequence_content_plot h5 {
  display: inline-block;
  height: 18px;
}
#fastqc_per_base_sequence_content_export_btn {
  font-family: "Lucida Grande", "Lucida Sans Unicode", Arial, Helvetica, sans-serif;
  float: right;
  color: #ccc;
  font-size: 12px;
  background: none;
  border: none;
  padding: 5px;
  border-radius: 5px;
  opacity: 0.9;
}
#fastqc_per_base_sequence_content_export_btn:hover,
#fastqc_per_base_sequence_content_export_btn:focus,
#fastqc_per_base_sequence_content_export_btn:active {
  background-color: #e6e6e6;
  color: #999;
}
#fastqc_per_base_sequence_content_export_btn span {
  margin-right: 5px;
  font-size: 14px;
  -webkit-transform: scale(1.3, 1);
  -moz-transform: scale(1.3, 1);
  -ms-transform: scale(1.3, 1);
  -o-transform: scale(1.3, 1);
  transform: scale(1.3, 1);
}

.fastqc_seq_heatmap_key {
  font-size: 12px;
}
.fastqc_seq_heatmap_key div {
  display: inline-block;
  width: 120px;
  margin: 0 8px;
}
#fastqc_seq_heatmap_key_pos {
  width: 80px;
}
.fastqc_seq_heatmap_key > span,
.fastqc_seq_heatmap_key div > span {
  display: inline-block;
  width: 60px;
}
#fastqc_seq_heatmap_key_t {
  border-bottom: 1px solid red;
}
#fastqc_seq_heatmap_key_c {
  border-bottom: 1px solid blue;
}
#fastqc_seq_heatmap_key_a {
  border-bottom: 1px solid green;
}
#fastqc_seq_heatmap_key_g {
  border-bottom: 1px solid black;
}
